#!/usr/bin/env python

# //******************************************************************************
# //
# //  rpn.py
# //
# //  RPN command-line calculator
# //  copyright (c) 2019, Rick Gutleber (rickg@his.com)
# //
# //  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
# //  information).
# //
# //******************************************************************************

#  https://en.wikipedia.org/wiki/Medieval_weights_and_measures

#  http://mathworld.wolfram.com/FavardConstants.html
#  How can we create a lambda to produce multiple Favard constants with a single expression?

#  rpn 4 meters meters * is_square

# Wind chill temperature index is a measurement of the combined cooling effect of low air temperature and
# wind on the human body. The index was first defined by the American Antarctic explorer Paul Siple in 1939.
# As originally used by U.S. meteorologists, the wind chill index (WCI) was computed from the temperature T
# (in F) and wind speed V (in mi/hr) using the formula:
# WCI = 0.0817(3.71 sqrt(V) + 5.81 - 0.25V)(T - 91.4) + 91.4.

# In 2001, U.S. and Canadian meteorologists agreed on a revised definition of the wind chill temperature
# index (WCTI). The U.S. version, with the temperature in F and wind speed in miles per hour, is
# WCTI = 35.74 + 0.6215T - 35.75V0.16 + 0.4275TV0.16.

# The Canadian formula, with temperatures in C and wind speed in km/h, is
# WCTI = 13.12 + 0.6215T - 11.37V0.16 + 0.3965TV0.16.

from __future__ import print_function

import six

import argparse
import os
import sys
import signal
import time

from contextlib import contextmanager
from mpmath import inf, mp, nan, nstr
from pathlib import Path

from rpn.rpnAliases import operatorAliases
from rpn.rpnConstantUtils import loadConstants
from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement

from rpn.rpnOperator import checkForVariable, RPNVariable

from rpn.rpnOperators import evaluateTerm, functionOperators, loadUnitNameData, \
                             loadUserFunctionsFile, preprocessTerms, RPNFunction, \
                             saveResult, saveUserFunctionsFile, setAccuracy, setPrecision

from rpn.rpnOutput import formatDateTime, formatListOutput, formatOutput, formatUnits, \
                          printHelp, printHelpModeHelp, printInteractiveHelp, printTitleScreen

from rpn.rpnPersistence import loadUnitData, loadUserVariablesFile, saveUserVariablesFile, \
                               loadUserConfigurationFile, saveUserConfigurationFile

from rpn.rpnPrimeUtils import checkForPrimeData

from rpn.rpnSpecial import handleIdentify

from rpn.rpnUtils import debugPrint, getCurrentArgList, getDataPath, \
                         parseNumerals, validateArguments, validateOptions

from rpn.rpnVersion import RPN_PROGRAM_NAME, PROGRAM_NAME, PROGRAM_VERSION, \
                           PROGRAM_DESCRIPTION, COPYRIGHT_MESSAGE

import rpn.rpnGlobals as g


# //******************************************************************************
# //
# //  lookAhead
# //
# //******************************************************************************

def lookAhead( iterable ):
    '''
    Pass through all values from the given iterable, augmented by the
    information if there are more values to come after the current one
    (True), or if it is the last value (False).
    '''
    # Get an iterator and pull the first value.
    it = iter( iterable )
    last = next( it )

    # Run the iterator to exhaustion (starting from the second value).
    for val in it:
        # Report the *previous* value (more to come).
        yield last, True
        last = val

    # Report the last value.
    yield last, False


# //******************************************************************************
# //
# //  evaluate
# //
# //******************************************************************************

def evaluate( terms ):
    '''
    This is the core function in rpn that evaluates the terms to be
    calculated.  terms are put into a stack, and popped off one at a time
    when evaluated.  When an operator is popped off and evaluated, the results
    get pushed back on to the stack for further processing, or ultimately,
    output.
    '''
    valueList = list( )
    index = 1                 # only used for error messages

    # handle a unit operator
    if not g.unitOperators:
        loadUnitData( )
        loadConstants( )

    # start parsing terms and populating the evaluation stack... this is the heart of rpn
    for term, hasMore in lookAhead( terms ):
        if term in g.operatorAliases:
            term = g.operatorAliases[ term ]

        if term in functionOperators:
            if g.creatingFunction:
                g.creatingFunction = False
            else:
                raise ValueError( 'function operators require a function definition' )

        currentValueList = getCurrentArgList( valueList )

        if g.creatingFunction:
            currentValueList[ -1 ].add( term )
            continue

        try:
            if not evaluateTerm( term, index, currentValueList, not hasMore ):
                valueList = [ nan ]
                break
        except ValueError as error:
            print( 'rpn:  error:  {0}'.format( error ) )
            valueList = [ nan ]
            break

        index = index + 1

    if len( g.echoArguments ) > 0:
        returnValue = list( g.echoArguments )

        if isinstance( valueList, list ) and len( valueList ) == 1:
            returnValue.append( valueList[ 0 ] )
        else:
            returnValue.append( valueList )

        g.echoArguments = [ ]
        return [ returnValue ]

    return valueList


# //******************************************************************************
# //
# //  handleOutput
# //
# //******************************************************************************

def handleOutput( valueList, indent=0, file=sys.stdout ):
    '''
    Once the evaluation of terms is complete, the results need to be
    translated into output.

    If the result is a list or a generator, special formatting turns those
    into text output.  Date-time values and measurements also require special
    formatting.

    Setting file to an io.StringIO objects allows for 'printing' to a string,
    which is used by makeHelp.py to generate actual rpn output for the examples.
    '''
    if valueList is None:
        return file

    indentString = ' ' * indent

    if len( valueList ) != 1:
        valueList = [ valueList ]

    if isinstance( valueList[ 0 ], RPNFunction ):
        print( indentString + 'rpn:  unexpected end of input in function definition', file=file )
    else:
        mp.pretty = True
        result = valueList.pop( )

        if result is nan:
            return file

        if g.comma:
            g.integerGrouping = 3     # override whatever was set on the command-line
            g.leadingZero = False     # this one, too
            g.integerDelimiter = ','
        else:
            g.integerDelimiter = ' '

        if isinstance( result, RPNGenerator ):
            formatListOutput( result.getGenerator( ), indent=indent, file=file )
        elif isinstance( result, list ):
            formatListOutput( result, indent=indent, file=file )
        else:
            if isinstance( result, RPNVariable ):
                result = result.getValue( )
            elif isinstance( result, RPNDateTime ):
                outputString = formatDateTime( result )
            elif isinstance( result, str ):
                result = checkForVariable( result )
                outputString = result
            else:
                # output the answer with all the extras according to command-line arguments

                # handle the units if we are displaying a measurement
                if isinstance( result, RPNMeasurement ):
                    outputString = formatOutput( nstr( result.getValue( ), mp.dps, min_fixed=-g.maximumFixed - 1 ) )
                    outputString += ' ' + formatUnits( result )
                else:
                    outputString = formatOutput( nstr( result, mp.dps, min_fixed=-g.maximumFixed - 1 ) )

            print( indentString + outputString, file=file )

            # handle --identify
            if g.identify:
                handleIdentify( result, file )

        saveResult( result )

    if g.timer or g.tempTimerMode:
        print( '\n' + indentString + '{:.3f} seconds'.format( time.process_time( ) - g.startTime ), file=file )


# //******************************************************************************
# //
# //  enterInteractiveMode
# //
# //******************************************************************************

def enterInteractiveMode( ):
    '''
    If rpn is launched with no expression, then it goes into interactive
    mode, where it will continue to evaluate new expressions input until
    the 'exit' command.
    '''
    try:
        import readline
    except ImportError:
        import pyreadline as readline

    readline.parse_and_bind( 'tab: complete' )
    readline.parse_and_bind( 'set editing-mode vi' )

    printTitleScreen( PROGRAM_NAME, PROGRAM_DESCRIPTION )

    g.results.append( None )   # g.results[ 0 ]
    g.interactive = True

    while True:
        g.promptCount += 1

        # clear single operation flags
        g.tempCommaMode = False
        g.tempHexMode = False
        g.tempIdentifyMode = False
        g.tempLeadingZeroMode = False
        g.tempOctalMode = False
        g.tempTimerMode = False

        try:
            line = input( 'rpn (' + str( g.promptCount ) + ')>' )
        except EOFError:
            break

        line = line.strip( )

        if line in [ 'exit', 'quit' ]:
            break

        terms = line.split( ' ' )

        if terms[ 0 ] == 'help':
            enterHelpMode( terms[ 1 : ] )
        else:
            if g.timer or g.tempTimerMode:
                g.startTime = time.process_time( )

            #newTerms = preprocessTerms( terms )
            #print( 'newTerms', newTerms )

            if validateArguments( terms ):
                valueList = evaluate( terms )

                g.results.append( valueList[ -1 ] )

                handleOutput( valueList )
            else:
                g.results.append( 0 )


# //******************************************************************************
# //
# //  enterHelpMode
# //
# //******************************************************************************

def enterHelpMode( terms ):
    '''
    When using rpn interactively, help is a special mode, which allows the user
    to navigate the help contents with much fewer keystrokes than having to
    invoke help over and over.
    '''
    printHelpModeHelp( )

    while True:
        try:
            line = input( 'rpn help>' )
        except EOFError:
            break

        line = line.strip( )

        if line in [ 'exit', 'quit' ]:
            break

        terms = line.split( ' ' )

        if terms[ 0 ] == 'help':
            printHelpModeHelp( )
        elif terms[ 0 ] == 'topics':
            printInteractiveHelp( )
        else:
            for term in terms:
                printHelp( term, True )


# //******************************************************************************
# //
# //  rpn
# //
# //******************************************************************************

def rpn( cmd_args ):
    '''
    This is the main function which processes the command-line arguments,
    handling both options and the expression to evaluate.   This function is
    mainly concerned with parsing and handling the command-line options.

    It finally calls evaluate( ) with the expression to be calculated, and
    returns the results, which can be formatted for output or used in another
    way (such as the unit test functionality).
    '''
    # initialize globals
    g.outputRadix = 10

    # look for help argument before we start setting everything up (because it's faster this way)
    help = False
    helpArg = ''

    for i in range( 0, len( cmd_args ) ):
        if cmd_args[ i ] == 'help':
            help = True
        else:
            if help:
                helpArg = cmd_args[ i ]

    if help:
        parser = argparse.ArgumentParser( prog = PROGRAM_NAME, description = RPN_PROGRAM_NAME +
                                          ' - ' + PROGRAM_DESCRIPTION + '\n    ' +
                                          COPYRIGHT_MESSAGE, add_help = False,
                                          formatter_class = argparse.RawTextHelpFormatter,
                                          prefix_chars = '-' )

        parser.add_argument( 'terms', nargs = '*', metavar = 'term' )
        parser.add_argument( '-l', '--line_length', type = int, action = 'store',
                             default = g.defaultLineLength )

        args = parser.parse_args( cmd_args )

        loadUnitNameData( )

        g.operatorAliases.update( operatorAliases )

        printHelp( helpArg )
        return

    # set up the command-line options parser
    parser = argparse.ArgumentParser( prog = PROGRAM_NAME, description = RPN_PROGRAM_NAME +
                                      ' - ' + PROGRAM_DESCRIPTION + '\n    ' +
                                      COPYRIGHT_MESSAGE, add_help = False,
                                      formatter_class = argparse.RawTextHelpFormatter,
                                      prefix_chars = '-' )

    parser.add_argument( '-a', '--output_accuracy', nargs = '?', type = int, action = 'store',
                         default = g.defaultOutputAccuracy, const = g.defaultOutputAccuracy )
    parser.add_argument( '-b', '--input_radix', type = str, action = 'store',
                         default = g.defaultInputRadix )
    parser.add_argument( '-c', '--comma', action = 'store_true' )
    parser.add_argument( '-d', '--decimal_grouping', nargs = '?', type = int, action = 'store',
                         default = 0, const = g.defaultDecimalGrouping )
    parser.add_argument( '-D', '--DEBUG', action = 'store_true' )
    parser.add_argument( '-e', '--profile', action = 'store_true' )
    parser.add_argument( '-E', '--echo_command', action = 'store_true' )
    parser.add_argument( '-g', '--integer_grouping', nargs = '?', type = int, action = 'store',
                         default = 0, const = g.defaultIntegerGrouping )
    parser.add_argument( '-h', '--help', action = 'store_true' )
    parser.add_argument( '-i', '--identify', action = 'store_true' )
    parser.add_argument( '-I', '--ignore_cache', action = 'store_true' )
    parser.add_argument( '-l', '--line_length', type = int, action = 'store',
                         default = g.defaultLineLength )
    parser.add_argument( '-m', '--maximum_fixed', type = int, action = 'store', default = g.defaultMaximumFixed )
    parser.add_argument( '-n', '--numerals', type = str, action = 'store', default = g.defaultNumerals )
    parser.add_argument( '-o', '--octal', action = 'store_true' )
    parser.add_argument( '-p', '--precision', type = int, action = 'store', default = g.defaultPrecision )
    parser.add_argument( '-r', '--output_radix', type = str, action = 'store',
                         default = g.defaultOutputRadix )
    parser.add_argument( '-s', '--list_format_level', nargs = '?', type = int, action = 'store', default = 0,
                         const = g.defaultListFormatLevel )
    parser.add_argument( '-t', '--timer', action = 'store_true' )
    parser.add_argument( '-T', '--time_limit', nargs = '?', type = int, action = 'store',
                         default = 0, const = g.timeLimit )
    parser.add_argument( '-V', '--version', action = 'store_true' )
    parser.add_argument( '-v', '--verbose', action = 'store_true' )
    parser.add_argument( '-w', '--bitwise_group_size', type = int, action = 'store',
                         default = g.defaultBitwiseGroupSize )
    parser.add_argument( '-x', '--hex', action = 'store_true' )
    parser.add_argument( '-z', '--leading_zero', action = 'store_true' )
    parser.add_argument( '-!', '--print_options', action = 'store_true' )
    parser.add_argument( '-?', '--other_help', action = 'store_true' )

    # pull out the options and the terms
    options = [ ]
    terms = [ ]

    loadUserVariablesFile( )
    loadUserFunctionsFile( )
    loadUserConfigurationFile( )

    if 'yafu_path' in g.userConfiguration and 'yafu_binary' in g.userConfiguration:
        g.useYAFU = True
    else:
        g.useYAFU = False

    for i, arg in enumerate( cmd_args ):
        if ( len( arg ) > 1 ) :
            if arg[ 0 ] == '$' and arg[ 1 : ] not in g.userVariables:
                raise ValueError( 'undefined user variable referenced: ' + arg )

            if arg[ 0 ] == '@' and arg[ 1 : ] not in g.userFunctions:
                raise ValueError( 'undefined user function referenced: ' + arg )

            if ( arg[ 0 ] == '-' ):
                if arg[ 1 ].isdigit( ):     # a negative number, not an option
                    terms.append( arg )
                else:
                    options.append( arg )
            else:
                terms.append( arg )
        else:
            terms.append( arg )

    debugPrint( 'terms', terms )
    debugPrint( 'options', options )

    # OK, let's parse and validate the options
    args = parser.parse_args( options )

    g.operatorAliases.update( operatorAliases )

    if args.help or args.other_help:
        loadUnitNameData( )

        printHelp( '' )
        return

    valid, errorString = validateOptions( args )

    if not valid:
        print( 'rpn:  ' + errorString )
        return

    # these are either globals or can be modified by other options (like -x)
    g.bitwiseGroupSize = args.bitwise_group_size
    g.integerGrouping = args.integer_grouping
    g.leadingZero = args.leading_zero

    # handle -a - set precision to be at least 2 greater than output accuracy
    setAccuracy( args.output_accuracy )

    # handle -b
    g.inputRadix = int( args.input_radix )

    # handle -c
    g.comma = args.comma

    # handle -d
    g.decimalGrouping = args.decimal_grouping

    # handle -D
    if args.DEBUG:
        g.debugMode = True

    # handle -e
    g.echo_command = args.echo_command

    # handle -i
    g.identify = args.identify

    # handle -I
    g.ignoreCache = args.ignore_cache

    # handle -l
    g.lineLength = args.line_length

    # handle -m
    g.maximumFixed = args.maximum_fixed

    # handle -n
    g.numerals = parseNumerals( args.numerals )

    # handle -o
    if args.octal:
        g.outputRadix = 8
        g.leadingZero = True
        g.integerGrouping = 3
        g.bitwiseGroupSize = 9

    # handle -p
    setPrecision( args.precision )

    # handle -r
    if args.output_radix == 'phi':
        g.outputRadix = g.phiBase
    elif args.output_radix == 'fib':
        g.outputRadix = g.fibBase
    elif args.output_radix == 'fac' or args.output_radix == '!':
        g.outputRadix = g.facBase
    elif args.output_radix == 'fac2' or args.output_radix == 'double_fac' or args.output_radix == '!!':
        g.outputRadix = g.doublefacBase
    elif args.output_radix == 'square' or args.output_radix == 'sqr':
        g.outputRadix = g.squareBase
    elif args.output_radix == 'lucas':
        g.outputRadix = g.lucasBase
    elif args.output_radix == 'triangular' or args.output_radix == 'tri':
        g.outputRadix = g.triangularBase
    elif args.output_radix == 'primorial':
        g.outputRadix = g.primorialBase
    elif args.output_radix == 'e':
        g.outputRadix = g.eBase
    elif args.output_radix == 'pi':
        g.outputRadix = g.piBase
    elif args.output_radix == 'sqrt2':
        g.outputRadix = g.sqrt2Base
    else:
        try:
            # if g.outputRadix was already set (e.g., by -o) then we don't want to override it
            if g.outputRadix == 10:
                g.outputRadix = int( args.output_radix )
        except ValueError:
            print( 'rpn:  can\'t interpret output radix \'%s\' as a number' % args.output_radix )
            return [ nan ]

    # -r validation
    if ( ( g.outputRadix < g.maxSpecialBase ) or ( g.outputRadix == 0 ) or
           ( g.outputRadix == 1 ) or ( g.outputRadix > 62 ) ):
        print( 'rpn:  output radix must be from 2 to 62, fib, phi, fac, doublefac, square, lucas' )
        return [ nan ]

    # handle -s
    g.listFormatLevel = args.list_format_level

    # handle -t
    g.timer = args.timer

    # handle -T
    g.timeLimit = args.time_limit

    # handle -v
    g.verbose = args.verbose

    # handle -V
    if args.version:
        return [ int( i ) for i in PROGRAM_VERSION.split( '.' ) ]

    # handle -x
    if args.hex:
        g.outputRadix = 16
        g.leadingZero = True
        g.integerGrouping = 4
        g.bitwiseGroupSize = 16

    # handle -u and -y:  mpmath wants precision of at least 53 for these functions
    if args.identify and mp.dps < 53:
        setAccuracy( 53 )

    if args.print_options:
        print( '--output_accuracy:  %d' % g.outputAccuracy )
        print( '--input_radix:  %d' % g.inputRadix )
        print( '--comma:  ' + ( 'true' if g.comma else 'false' ) )
        print( '--decimal_grouping:  %d' % g.decimalGrouping )
        print( '--integer_grouping:  %d' % g.integerGrouping )
        print( '--line_length:  %d' % g.lineLength )
        print( '--numerals:  ' + g.numerals )
        print( '--octal:  ' + ( 'true' if args.octal else 'false' ) )
        print( '--precision:  %d' % args.precision )
        print( '--output_radix:  %d' % g.outputRadix )
        print( '--list_format_level:  %d' % g.listFormatLevel )
        print( '--timer:  ' + ( 'true' if args.timer else 'false' ) )
        print( '--verbose:  ' + ( 'true' if g.verbose else 'false' ) )
        print( '--bitwise_group_size:  %d' % g.bitwiseGroupSize )
        print( '--hex:  ' + ( 'true' if args.hex else 'false' ) )
        print( '--identify:  ' + ( 'true' if args.identify else 'false' ) )
        print( '--leading_zero:  ' + ( 'true' if g.leadingZero else 'false' ) )
        print( '--ignore_cache:  ' + ( 'true' if g.ignoreCache else 'false' ) )
        print( )

    g.creatingFunction = False

    # enter interactive mode if there are no arguments
    if not terms:
        if not loadUnitNameData( ):
            return

        enterInteractiveMode( )
        return

    # let's check out the arguments before we start to do any calculations
    if not validateArguments( terms ):
        return

    #newTerms = preprocessTerms( terms )
    #print( 'newTerms', newTerms )

    # waiting until we've validated the arguments to do this because it's slow
    if not loadUnitNameData( ):
        return

    if g.echo_command:
        print( *sys.argv )

    if g.timer:
        g.startTime = time.process_time( )

    return evaluate( terms )


# //******************************************************************************
# //
# //  __main__
# //
# //******************************************************************************

def main( ):
    checkForPrimeData( )

    unitsFile = Path( getDataPath( ) + os.sep + 'units.pckl.bz2' )

    if not unitsFile.is_file( ):
        print( 'Please run "makeUnits" to initialize the unit conversion data files.' )
        sys.exit( 0 )

    helpFile = Path( getDataPath( ) + os.sep + 'help.pckl.bz2' )

    if not helpFile.is_file( ):
        print( 'Please run "makeHelp" to initialize the help files.' )
        sys.exit( 0 )

    try:
        #for arg in sys.argv:
        #    if arg == '-e':
        #        import profile
        #        profile.run( 'handleOutput( rpn( sys.argv[ 1 : ] ) )' )

        handleOutput( rpn( sys.argv[ 1 : ] ) )

        if g.userVariablesAreDirty:
            saveUserVariablesFile( )

        if g.userFunctionsAreDirty:
            saveUserFunctionsFile( )

        if g.userConfigurationIsDirty:
            saveUserConfigurationFile( )
    except ValueError as error:
        print( '\nrpn:  value error:  {0}'.format( error ) )

        if g.debugMode:
            raise
        else:
            sys.exit( 0 )
    except KeyboardInterrupt:
        print( )
        print( 'handling ctrl-c keyboard interrupt...' )
        sys.exit( 0 )


# //******************************************************************************
# //
# //  __main__
# //
# //******************************************************************************

if __name__ == '__main__':
    main( )




#!/usr/bin/env python

# //******************************************************************************
# //
# //  rpnGlobals.py
# //
# //  RPN command-line calculator global declarations
# //  copyright (c) 2019, Rick Gutleber (rickg@his.com)
# //
# //  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
# //  information).
# //
# //******************************************************************************

# general globals
dataDir = 'rpndata'
ecm = True

# base identifiers
phiBase = -1
fibBase = -2
facBase = -3
doublefacBase = -4
squareBase = -5
lucasBase = -6
triangularBase = -7
primorialBase = -8
eBase = -9
piBase = -10
sqrt2Base = -11

maxSpecialBase = -11

# default options
defaultBitwiseGroupSize = 16
defaultCFTerms = 10
defaultDecimalDelimiter = ' '
defaultDecimalGrouping = 5
defaultInputRadix = 10
defaultIntegerGrouping = 3
defaultLineLength = 80
defaultListFormatLevel = 1
defaultMinValueForYAFU = 1000000000000000000
defaultMaximumFixed = 5
defaultOutputAccuracy = 12
defaultOutputRadix = 10
defaultPrecision = 20

defaultNumerals = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

# state variables
astroDataLoaded = False
astroDataAvailable = False
creatingFunction = False
duplicateOperations = 0
echoArguments = [ ]
helpLoaded = False
interactive = False
lastOperand = 0
nestedListLevel = 0
operandsToRemove = 0
operatorList = False
operatorsInList = 0
startTime = 0
useMembers = 0

# options
bitwiseGroupSize = 0
comma = False
debugMode = False
decimalDelimiter = defaultDecimalDelimiter
decimalGrouping = 0
echo_command = False
identify = False
ignoreCache = False
inputRadix = defaultInputRadix
leadingZero = False
lineLength = defaultLineLength
listFormatLevel = 0
maximumFixed = defaultMaximumFixed
maxToFactorByTrialDivision = 100000000  # 10,000 squared because we check all primes below 10,000
minValueForYAFU = defaultMinValueForYAFU
numerals = ''
outputAccuracy = -1
outputRadix = defaultOutputRadix
tempCommaMode = False
tempHexMode = False
tempIdentifyMode = False
tempLeadingZeroMode = False
tempOctalMode = False
tempTimerMode = False
timeLimit = 0
timer = False
useYAFU = True
verbose = False
zhangConjecturesAllowed = False

# unit data
basicUnitTypes = { }
constantOperators = { }
constantOperatorNames = [ ]
operatorAliases = { }
unitConversionMatrix = { }
unitOperators = { }
unitOperatorNames = [ ]

# help data
helpTopics = { }
helpVersion = "0.0.0"
operatorCategories = [ ]
operatorHelp = { }
unitTypeDict = { }

# interactive mode
results = [ ]
variables = { }
promptCount = 0

# factor cache
minValueToCache = 100000
factorCache = None

# location cache
locationCache = None

# user-defined constants
constants = { }

# all keywords
keywords = [ ]

# operator caches
functionCaches = { }

# prime number data
databases = { }
cursors = { }
primeDataAvailable = False

# chemistry tables
elements = None
atomic_numbers = None

# config
userVariables = { }
userVariablesAreDirty = False
userFunctions = { }
userFunctionsAreDirty = False
userConfiguration = { }
userConfigurationIsDirty = False

# astronomy data
planets = None
ephemeris = None
timescale = None

# constants
c = None
e0 = None
h_bar = None
G = None




#!/usr/bin/env python

# //******************************************************************************
# //
# //  makeUnits
# //
# //  RPN command-line calculator unit conversion data generator
# //  copyright (c) 2019, Rick Gutleber (rickg@his.com)
# //
# //  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
# //  information).
# //
# //******************************************************************************

from __future__ import print_function

import six

import bz2
import contextlib
import itertools
import os
import pickle
import time

from mpmath import almosteq, mp, fdiv, fmul

#  This has to go here so the mpf's in the import get created with 50 places of precision.
mp.dps = 50

from rpn.rpnConstantOperators import *
from rpn.rpnMeasurement import specialUnitConversionMatrix
from rpn.rpnUnits import *
from rpn.rpnUtils import getDataPath
from rpn.rpnUnitTypes import basicUnitTypes
from rpn.rpnVersion import PROGRAM_VERSION, PROGRAM_VERSION_STRING, COPYRIGHT_MESSAGE

import rpn.rpnGlobals as g


# //******************************************************************************
# //
# //  constants
# //
# //******************************************************************************

PROGRAM_NAME = 'makeUnits'
PROGRAM_DESCRIPTION = 'RPN command-line calculator unit conversion data generator'


# //******************************************************************************
# //
# //  makeMetricUnit
# //
# //******************************************************************************

def makeMetricUnit( prefix, unit ):
    # special case because the standard is inconsistent
    if ( unit == 'ohm' ) and ( prefix == 'giga' ):
        return 'gigaohm'   # not gigohm
    elif ( unit[ 0 ] == 'o' ) and ( prefix[ -1 ] in 'oa' ):
        return prefix[ : -1 ] + unit
    elif unit[ 0 ] == 'a' and ( ( prefix[ -1 ] == 'a' ) or ( prefix[ -3 : ] == 'cto' ) ):
        return prefix[ : -1 ] + unit
    else:
        return prefix + unit


# //******************************************************************************
# //
# //  makeUnitTypeTable
# //
# //  maps each unit type to a list of units with that type
# //
# //******************************************************************************

def makeUnitTypeTable( unitOperators ):
    unitTypeTable = { }

    for unitType in basicUnitTypes:
        unitTypeTable[ unitType ] = [ ]

    for unit in unitOperators:
        unitTypeTable[ unitOperators[ unit ].unitType ].append( unit )

    return unitTypeTable


# //******************************************************************************
# //
# //  makeAliases
# //
# //******************************************************************************

def makeAliases( ):
    newAliases = { }

    for metricUnit in metricUnits:
        for prefix in metricPrefixes:
            unit = makeMetricUnit( prefix[ 0 ], metricUnit )
            pluralUnit = makeMetricUnit( prefix[ 0 ], metricUnits[ metricUnit ][ 0 ] )

            if pluralUnit != unit:
                newAliases[ pluralUnit ] = unit             # add plural alias

            for alternateUnit in metricUnits[ metricUnit ][ 2 ]:           # add alternate spelling alias
                newAliases[ makeMetricUnit( prefix[ 0 ], alternateUnit ) ] = unit

            for alternateUnit in metricUnits[ metricUnit ][ 3 ]:           # add alternate spelling plural alias
                newAliases[ makeMetricUnit( prefix[ 0 ], alternateUnit ) ] = unit

    for dataUnit in dataUnits:
        newAliases[ dataUnit[ 2 ] ] = dataUnit[ 0 ]

        for prefix in dataPrefixes:
            unit = prefix[ 0 ] + dataUnit[ 0 ]
            pluralUnit = prefix[ 0 ] + dataUnit[ 1 ]

            if pluralUnit != unit:
                newAliases[ pluralUnit ] = unit                 # add plural alias

            newAliases[ prefix[ 1 ] + dataUnit[ 0 ] ] = unit    # add SI abbreviation alias
            newAliases[ prefix[ 1 ] + dataUnit[ 1 ] ] = unit    # add SI abbreviation alias
            newAliases[ prefix[ 1 ] + dataUnit[ 2 ] ] = unit    # add SI abbreviation alias

            for alternateUnit in dataUnit[ 3 ]:                 # add alternate spelling alias
                newAliases[ prefix[ 0 ] + alternateUnit ] = unit

            for alternateUnit in dataUnit[ 4 ]:                 # add alternate spelling plural alias
                newAliases[ prefix[ 0 ] + alternateUnit ] = unit

        for prefix in binaryPrefixes:
            unit = prefix[ 0 ] + dataUnit[ 0 ]
            pluralUnit = prefix[ 0 ] + dataUnit[ 1 ]

            if pluralUnit != unit:
                newAliases[ pluralUnit ] = unit                 # add plural alias

            newAliases[ prefix[ 1 ] + dataUnit[ 0 ] ] = unit    # add SI abbreviation alias
            newAliases[ prefix[ 1 ] + dataUnit[ 1 ] ] = unit    # add SI abbreviation alias
            newAliases[ prefix[ 1 ] + dataUnit[ 2 ] ] = unit    # add SI abbreviation alias

            for alternateUnit in dataUnit[ 3 ]:                 # add alternate spelling alias
                newAliases[ prefix[ 0 ] + alternateUnit ] = unit

            for alternateUnit in dataUnit[ 4 ]:                 # add alternate spelling plural alias
                newAliases[ prefix[ 0 ] + alternateUnit ] = unit

    for unit in unitOperators:
        unitInfo = unitOperators[ unit ]

        if unitInfo.plural != unit:
            newAliases[ unitInfo.plural ] = unit

        for alias in unitInfo.aliases:
            newAliases[ alias ] = unit

        if unitInfo.abbrev != '':
            newAliases[ unitInfo.abbrev ] = unit

    for constant in constantOperators:
        constantInfo = constantOperators[ constant ]

        for alias in constantInfo.aliases:
            newAliases[ alias ] = constant

    return newAliases


# //******************************************************************************
# //
# //  expandMetricUnits
# //
# //  Every metric unit needs to be permuted for all SI power types.  We need to
# //  create conversions for each new type, as well as aliases.
# //
# //******************************************************************************

def expandMetricUnits( ):
    # expand metric measurements for all prefixes
    metricConversions = { }
    metricAliases = { }

    for metricUnit in metricUnits:
        for prefix in metricPrefixes:
            newName = makeMetricUnit( prefix[ 0 ], metricUnit )
            newPlural = makeMetricUnit( prefix[ 0 ], metricUnits[ metricUnit][ 0 ] )

            if newName not in unitOperators:
                # constuct unit operator info
                helpText = '\nfill me out for metric units'

                if metricUnits[ metricUnit ][ 1 ]:
                    unitOperators[ newName ] = \
                        RPNUnitInfo( unitOperators[ metricUnit ].unitType, newName, newPlural,
                                     prefix[ 1 ] + metricUnits[ metricUnit ][ 1 ], [ ], [ 'SI' ],
                                     helpText, True )
                else:
                    unitOperators[ newName ] = \
                        RPNUnitInfo( unitOperators[ metricUnit ].unitType, newName, newPlural,
                                     helpText, [ ], [ 'SI' ], '', True )


                newConversion = power( 10, mpmathify( prefix[ 2 ] ) )
                unitConversionMatrix[ ( newName, metricUnit ) ] = newConversion
                newConversion = fdiv( 1, newConversion )
                unitConversionMatrix[ ( metricUnit, newName ) ] = newConversion
            else:
                newConversion = power( 10, fneg( mpmathify( prefix[ 2 ] ) ) )

            # create area and volume operators for new length units
            if unitOperators[ metricUnit ].unitType == 'length':
                # create new area operators
                newUnit = newName + '^2'

                areaConversion = power( newConversion, 2 )
                oldUnit = metricUnit + '^2'

                if newUnit not in unitOperators:
                    newUnitInfo, newUnitAliases = makeAreaOperator( newName, newPlural )
                    metricAliases.update( newUnitAliases )

                    unitOperators[ newUnit ] = newUnitInfo

                    # add new conversions
                    metricConversions[ ( oldUnit, newUnit ) ] = areaConversion
                    metricConversions[ ( newUnit, oldUnit ) ] = fdiv( 1, areaConversion )

                # create new volume operators
                newUnit = newName + '^3'

                volumeConversion = power( newConversion, 3 )
                oldUnit = metricUnit + '^3'

                if newUnit not in unitOperators:
                    newUnitInfo, newUnitAliases = makeVolumeOperator( newName, newPlural )
                    metricAliases.update( newUnitAliases )

                    unitOperators[ newUnit ] = newUnitInfo

                    # add new conversions
                    metricConversions[ ( oldUnit, newUnit ) ] = volumeConversion
                    metricConversions[ ( newUnit, oldUnit ) ] = fdiv( 1, volumeConversion )

    return metricConversions, metricAliases


# //******************************************************************************
# //
# //  expandDataUnits
# //
# //  Every data unit needs to be permuted for all positive SI power types and
# //  the binary power types.  We need to create conversions for each new type,
# //  as well as aliases.
# //
# //******************************************************************************

def expandDataUnits( ):
    # expand data measurements for all prefixes
    newConversions = { }

    for dataUnit in dataUnits:
        for prefix in dataPrefixes:
            newName = prefix[ 0 ] + dataUnit[ 0 ]
            newPlural = prefix[ 0 ] + dataUnit[ 1 ]

            # constuct unit operator info
            helpText = '\nfill me out for data units'

            unitOperators[ newName ] = \
                RPNUnitInfo( unitOperators[ dataUnit[ 0 ] ].unitType, newName, newPlural,
                             prefix[ 1 ] + dataUnit[ 2 ], [ ],
                             unitOperators[ dataUnit[ 0 ] ].categories, helpText, True )

            # create new conversions
            newConversion = power( 10, mpmathify( prefix[ 2 ] ) )
            newConversions[ ( newName, dataUnit[ 0 ] ) ] = newConversion
            newConversion = fdiv( 1, newConversion )
            newConversions[ ( dataUnit[ 0 ], newName ) ] = newConversion

        for prefix in binaryPrefixes:
            newName = prefix[ 0 ] + dataUnit[ 0 ]
            newPlural = prefix[ 0 ] + dataUnit[ 1 ]

            # constuct unit operator info
            helpText = '\nfill me out for binary data units'

            unitOperators[ newName ] = \
                RPNUnitInfo( unitOperators[ dataUnit[ 0 ] ].unitType, newName, newPlural,
                             prefix[ 1 ] + dataUnit[ 2 ], [ ],
                             unitOperators[ dataUnit[ 0 ] ].categories, helpText, True )

            # create new conversions
            newConversion = power( 2, mpmathify( prefix[ 2 ] ) )
            newConversions[ ( newName, dataUnit[ 0 ] ) ] = newConversion
            newConversion = fdiv( 1, newConversion )
            newConversions[ ( dataUnit[ 0 ], newName ) ] = newConversion

    return newConversions


# //******************************************************************************
# //
# //  makeAreaOperator
# //
# //******************************************************************************

def makeAreaOperator( unit, unitPlural ):
    unitInfo = unitOperators[ unit ]

    newAliases = { }

    newUnit = unit + '^2'

    if unitInfo.abbrev == '':
        abbrev = 'sq' + unit
    else:
        abbrev = 'sq' + unitInfo.abbrev
        newAliases[ 'sq' + unitInfo.abbrev ] = newUnit
        newAliases[ 'sq_' + unitInfo.abbrev ] = newUnit
        newAliases[ 'square_' + unitInfo.abbrev ] = newUnit

    newAliases[ 'square_' + unit ] = newUnit
    newAliases[ 'square_' + unitInfo.plural ] = newUnit
    newAliases[ 'sq_' + unit ] = newUnit
    newAliases[ 'sq_' + unitInfo.plural ] = newUnit
    newAliases[ 'sq' + unit ] = newUnit
    newAliases[ 'sq' + unitInfo.plural ] = newUnit

    helpText = '\nfill me out for area operator'

    newUnitInfo = RPNUnitInfo( 'area', newUnit, newUnit, abbrev, [ ],
                               unitInfo.categories, helpText, True )

    return newUnitInfo, newAliases


# //******************************************************************************
# //
# //  makeVolumeOperator
# //
# //******************************************************************************

def makeVolumeOperator( unit, unitPlural ):
    unitInfo = unitOperators[ unit ]

    newAliases = { }

    newUnit = unit + '^3'

    if unitInfo.abbrev == '':
        abbrev = 'cu' + unit
    else:
        abbrev = 'cu' + unitInfo.abbrev
        newAliases[ 'cu' + unitInfo.abbrev ] = newUnit
        newAliases[ 'cu_' + unitInfo.abbrev ] = newUnit
        newAliases[ 'cubic_' + unitInfo.abbrev ] = newUnit

    newAliases[ 'cubic_' + unit ] = newUnit
    newAliases[ 'cubic_' + unitInfo.plural ] = newUnit
    newAliases[ 'cu_' + unit ] = newUnit
    newAliases[ 'cu_' + unitInfo.plural ] = newUnit
    newAliases[ 'cu' + unit ] = newUnit
    newAliases[ 'cu' + unitInfo.plural ] = newUnit

    helpText = '\nfill me out for volume operator'

    newUnitInfo = RPNUnitInfo( 'volume', newUnit, newUnit, abbrev, [ ],
                               unitInfo.categories, helpText, True )

    return newUnitInfo, newAliases


# //******************************************************************************
# //
# //  createAreaAndVolumeOperators
# //
# //******************************************************************************

def createAreaAndVolumeOperators( unitOperators ):
    newOperators = { }
    newAliases = { }

    for unit in unitOperators:
        unitInfo = unitOperators[ unit ]

        if unitInfo.representation != unit:
            newAliases[ unitInfo.representation ] = unit

        if unitInfo.unitType == 'length':
            newUnit = unit + '^2'

            if newUnit not in unitOperators:
                newUnitInfo, newUnitAliases = \
                    makeAreaOperator( unit, unitOperators[ unit ].plural )

                newAliases.update( newUnitAliases )
                newOperators[ newUnit ] = newUnitInfo

                compoundUnit = unit + '^2'
                newAliases[ compoundUnit ] = newUnit

            newUnit = unit + '^3'

            if newUnit not in unitOperators:
                newUnitInfo, newUnitAliases = \
                    makeVolumeOperator( unit, unitOperators[ unit ].plural )

                newAliases.update( newUnitAliases )
                newOperators[ newUnit ] = newUnitInfo

                compoundUnit = unit + '^3'
                newAliases[ compoundUnit ] = newUnit

    return newOperators, newAliases


# //******************************************************************************
# //
# //  expandCompoundTimeUnits
# //
# //******************************************************************************

def expandCompoundTimeUnits( unitConversionMatrix, unitOperators, newAliases ):
    newUnitOperators = { }

    # We need to store the new ones in a different dictionary because we can't
    # modified unitOperators while iterating through it.  This is only guaranteed
    # a single unit compounded with seconds, but that's all I'm interested in.
    for unit in unitOperators:
        unitRoot = ''

        if '/' in unit or '^' in unit:
            continue

        originalUnitInfo = unitOperators[ unit ]

        if unit[ -7 : ] == '*second':
            unitRoot = unit[ : -7 ]

        if unit[ : 7 ] == 'second*':
            unitRoot = unit[ 7 : ]

        if unitRoot:
            newRoots = [ unitRoot ]

            if unitRoot in metricUnits:
                for prefix in metricPrefixes:
                    newRoots.append( prefix[ 0 ] + unitRoot )

            for newRoot in newRoots:
                if newRoot not in unitOperators:
                    continue

                unitInfo = unitOperators[ newRoot ]
                rootUnitInfo = unitOperators[ newRoot ]

                for timeUnit in timeUnits:
                    newUnit = newRoot + '*' + timeUnit[ 0 ]

                    # make a bunch of obvious aliases
                    newAliases[ newRoot + '-' + timeUnit[ 0 ] ] = newUnit
                    newPlural = newRoot + '-' + timeUnit[ 1 ]
                    newAliases[ newPlural ] = newUnit
                    newAliases[ newRoot + '-' + timeUnit[ 1 ] ] = newUnit

                    # We assume the abbrev ends with an s for second
                    if unitInfo.abbrev != '':
                        newAbbrev = unitInfo.abbrev + timeUnit[ 2 ]
                        newAliases[ newAbbrev ] = newUnit

                    for alias in rootUnitInfo.aliases:
                        newAliases[ alias + '*' + timeUnit[ 0 ] ] = newUnit
                        newAliases[ alias + '-' + timeUnit[ 0 ] ] = newUnit

                        if timeUnit[ 0 ] != timeUnit[ 1 ]:
                            newAliases[ alias + '*' + timeUnit[ 1 ] ] = newUnit
                            newAliases[ alias + '-' + timeUnit[ 1 ] ] = newUnit

                    helpText = '\nfill me out for compound time units'

                    newUnitOperators[ newUnit ] = \
                        RPNUnitInfo( originalUnitInfo.unitType, newRoot + '*' + timeUnit[ 0 ], newPlural,
                                     '', [ ], originalUnitInfo.categories, helpText, True )

                    conversion = mpmathify( timeUnit[ 3 ] )
                    unitConversionMatrix[ ( newUnit, unit ) ] = conversion
                    unitConversionMatrix[ ( unit, newUnit ) ] = fdiv( 1, conversion )

                    #print( 'newUnit', newUnit )
                    #print( 'unit', unit )
                    #print( 'conversion', conversion )
                    #print( )

    unitOperators.update( newUnitOperators )


# //******************************************************************************
# //
# //  extrapolateTransitiveConversions
# //
# //******************************************************************************

def extrapolateTransitiveConversions( op1, op2, unitTypeTable, unitType, unitConversionMatrix ):
    newConversions = { }

    for op3 in unitTypeTable[ unitType ]:
        # we can ignore duplicate operators
        if op3 in [ op1, op2 ]:
            continue

        conversion = unitConversionMatrix[ ( op1, op2 ) ]

        if ( op1, op3 ) not in unitConversionMatrix and \
           ( op2, op3 ) in unitConversionMatrix:
            # print( 'transitive: ', ( op2, op3 ),
            #        unitConversionMatrix[ ( op2, op3 ) ] )
            newConversion = fmul( conversion, unitConversionMatrix[ ( op2, op3 ) ] )
            # print( ( op1, op3 ), newConversion )
            newConversions[ ( op1, op3 ) ] = newConversion
            # print( ( op3, op1 ), fdiv( 1, newConversion ) )
            newConversions[ ( op3, op1 ) ] = fdiv( 1, newConversion )
        elif ( op2, op3 ) not in unitConversionMatrix and \
             ( op1, op3 ) in unitConversionMatrix:
            # print( 'transitive: ', ( op1, op3 ),
            #        unitConversionMatrix[ ( op1, op3 ) ] )
            newConversion = fdiv( unitConversionMatrix[ ( op1, op3 ) ], conversion )
            # print( ( op2, op3 ), newConversion )
            newConversions[ ( op2, op3 ) ] = newConversion
            # print( ( op3, op2 ), fdiv( 1, newConversion ) )
            newConversions[ ( op3, op2 ) ] = fdiv( 1, newConversion )

    return newConversions


# //******************************************************************************
# //
# //  testAllCombinations
# //
# //  Let's make sure all the conversions exist (except for transitive conversions
# //  involving units that use the special unit conversion matrix.
# //
# //  This means this test will print out some errors regardless.
# //
# //******************************************************************************

def testAllCombinations( unitTypeTable, unitConversionMatrix ):
    for unitType in unitTypeTable:
        print( unitType, '*************************************************' )

        for unit1, unit2 in itertools.combinations( unitTypeTable[ unitType ], 2 ):
            if ( unit1, unit2 ) not in unitConversionMatrix and \
               ( unit1, unit2 ) not in specialUnitConversionMatrix:
                print( 'conversion not found for', unit1, 'and', unit2 )


# //******************************************************************************
# //
# //  testAllConversions
# //
# //  Let's prove the conversions are consistent.
# //
# //******************************************************************************

def testAllConversions( unitTypeTable, unitConversionMatrix ):
    for unitType in unitTypeTable:
        for unit1, unit2, unit3 in itertools.permutations( unitTypeTable[ unitType ], 3 ):
            try:
                factor1 = unitConversionMatrix[ unit1, unit2 ]
                factor2 = unitConversionMatrix[ unit2, unit3 ]
                factor3 = unitConversionMatrix[ unit1, unit3 ]
            except:
                continue

            if not almosteq( fmul( factor1, factor2 ), factor3, rel_eps=1.0e-10 ):
                print( 'conversion inconsistency found for ' + unit1 + ', ' + unit2 + ', and', unit3 )
                print( factor1, factor2, factor3 )
                print( fmul( factor1, factor2 ), factor3 )
                print( )


# //******************************************************************************
# //
# //  initializeConversionMatrix
# //
# //******************************************************************************

def initializeConversionMatrix( unitConversionMatrix ):
    # reverse each conversion
    print( 'Reversing each conversion...' )

    newConversions = { }

    for op1, op2 in unitConversionMatrix:
        conversion = fdiv( 1, unitConversionMatrix[ ( op1, op2 ) ] )
        newConversions[ ( op2, op1 ) ] = conversion

    unitConversionMatrix.update( newConversions )

    # create area and volume units from all of the length units
    print( 'Creating area and volume units for all length units...' )

    newOperators, newAliases = createAreaAndVolumeOperators( unitOperators )

    unitOperators.update( newOperators )

    # add new conversions for the new area and volume units
    print( 'Adding new conversions for the new area and volume units...' )

    newConversions = { }

    for op1, op2 in unitConversionMatrix:
        if unitOperators[ op1 ].unitType == 'length':
            conversion = unitConversionMatrix[ ( op1, op2 ) ]
            newConversions[ ( op1 + '^2', op2 + '^2' ) ] = power( conversion, 2 )
            newConversions[ ( op1 + '^3', op2 + '^3' ) ] = power( conversion, 3 )

    unitConversionMatrix.update( newConversions )

    print( 'Expanding metric units against the list of SI prefixes...' )

    metricConversions, metricAliases = expandMetricUnits( )

    unitConversionMatrix.update( metricConversions )
    newAliases.update( metricAliases )

    print( 'Expanding data units against the list of SI and binary prefixes...' )
    unitConversionMatrix.update( expandDataUnits( ) )

    # add new operators for compound time units
    print( 'Expanding compound time units...' )

    expandCompoundTimeUnits( unitConversionMatrix, unitOperators, newAliases )

    # extrapolate transitive conversions
    print( )
    print( 'Extrapolating transitive conversions for', len( unitOperators ), 'units...' )

    unitTypeTable = makeUnitTypeTable( unitOperators )

    for unitType in sorted( basicUnitTypes ):
        if unitType != '_null_type':
            print( '\r     ', unitType, '({} units)'.format( len( unitTypeTable[ unitType ] ) ) )

        while True:
            newConversions = { }

            for op1, op2 in itertools.combinations( unitTypeTable[ unitType ], 2 ):
                if ( op1, op2 ) in unitConversionMatrix:
                    # print( )
                    # print( ( op1, op2 ), ': ', unitConversionMatrix[ ( op1, op2 ) ] )
                    newConversions.update( \
                        extrapolateTransitiveConversions( op1, op2, unitTypeTable, unitType, \
                                                          unitConversionMatrix ) )

            if not newConversions:
                break

            unitConversionMatrix.update( newConversions )

            print( '\r' + str( len( unitConversionMatrix ) ), end='' )

    print( '\r' + str( len( unitConversionMatrix ) ) + ' conversions' )

    # make some more aliases
    print( '        ' )
    print( 'Making some more aliases...' )

    newAliases.update( makeAliases( ) )

    print( 'Stringifying conversion matrix values...' )
    for op1, op2 in unitConversionMatrix:
        unitConversionMatrix[ ( op1, op2 ) ] = str( unitConversionMatrix[ ( op1, op2 ) ] )

    print( 'Saving everything...' )

    # save the list of unit operator names and aliases
    fileName = getDataPath( ) + os.sep + 'unit_names.pckl.bz2'

    with contextlib.closing( bz2.BZ2File( fileName, 'wb' ) ) as pickleFile:
        pickle.dump( PROGRAM_VERSION, pickleFile )
        pickle.dump( list( unitOperators.keys( ) ), pickleFile )
        pickle.dump( list( constantOperators.keys( ) ), pickleFile )
        pickle.dump( newAliases, pickleFile )

    # save the actual unit data
    fileName = getDataPath( ) + os.sep + 'units.pckl.bz2'

    #testAllCombinations( unitTypeTable, unitConversionMatrix )
    #testAllConversions( unitTypeTable, unitConversionMatrix )

    with contextlib.closing( bz2.BZ2File( fileName, 'wb' ) ) as pickleFile:
        pickle.dump( PROGRAM_VERSION, pickleFile )
        pickle.dump( basicUnitTypes, pickleFile )
        pickle.dump( unitOperators, pickleFile )
        pickle.dump( constantOperators, pickleFile )

    fileName = getDataPath( ) + os.sep + 'unit_conversions.pckl.bz2'

    with contextlib.closing( bz2.BZ2File( fileName, 'wb' ) ) as pickleFile:
        pickle.dump( unitConversionMatrix, pickleFile )

    unitTypeDict = { }

    for unitType in basicUnitTypes.keys( ):
        unitTypeDict[ unitType ] = list( )

    for unit in unitOperators:
        unitTypeDict[ unitOperators[ unit ].unitType ].append( unit )

    fileName = getDataPath( ) + os.sep + 'unit_help.pckl.bz2'

    with contextlib.closing( bz2.BZ2File( fileName, 'wb' ) ) as pickleFile:
        pickle.dump( unitTypeDict, pickleFile )

    print( )
    print( '{:,} unit operators'.format( len( unitOperators ) ) )
    print( '{:,} unit conversions'.format( len( unitConversionMatrix ) ) )
    print( '{:,} aliases'.format( len( newAliases ) ) )


# //******************************************************************************
# //
# //  main
# //
# //******************************************************************************

def main( ):
    print( PROGRAM_NAME + PROGRAM_VERSION_STRING + ' - ' + PROGRAM_DESCRIPTION )
    print( COPYRIGHT_MESSAGE )
    print( )

    startTime = time.process_time( )

    initializeConversionMatrix( unitConversionMatrix )

    print( )
    print( 'Unit data completed.  Time elapsed:  {:.3f} seconds'.format( time.process_time( ) - startTime ) )


# //******************************************************************************
# //
# //  __main__
# //
# //******************************************************************************

if __name__ == '__main__':
    main( )

